/*
 * findtemp.c
 *
 * Scan the file access policy and display nonexistent path.
 *
 * Copyright (C) 2005-2006  NTT DATA CORPORATION
 *
 * Version: 1.1   2006/04/01
 *
 * This program checks the existence for pathnames which have the write permission,
 * and displays the pathname if nonexistent, for such pathnames are likely temporary.
 *
 * This program can also check the existence for all pathnames,
 * and displays the pathname if nonexistent.
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

static int decode(const char *ascii, char *bin) {
	char c, d, e;
	while ((c = *bin++ = *ascii++) != '\0') {
		if (c == '\\') {
			c = *ascii++;
			switch (c) {
			case '\\':      /* "\\" */
				continue;
			case '0':       /* "\ooo" */
			case '1':
			case '2':
			case '3':
				if ((d = *ascii++) >= '0' && d <= '7' && (e = *ascii++) >= '0' && e <= '7') {
					const unsigned char f =
						(((unsigned char) (c - '0')) << 6) +
						(((unsigned char) (d - '0')) << 3) +
						(((unsigned char) (e - '0')));
					if (f && (f <= ' ' || f >= 127)) {
						*(bin - 1) = f;
						continue; /* pattern is not \000 */
					}
				}
			}
			return 0;
		} else if (c <= ' ' || c >= 127) {
			return 0;
		}
	}
	return 1;
}

int main(int argc, char *argv[]) {
	const char **pattern_list = NULL;
	int pattern_list_count = 0;
	int i, j;
	int detect_all = 0;
	static char buffer[16384];
	if (argc > 1) {
		if (strcmp(argv[1], "--all") == 0) {
			detect_all = 1;
		} else {
			printf("%s [--all] < domain_policy\n\n", argv[0]);
			printf("If --all is given, all nonexistent pathnames are shown.\n");
			printf("If --all is not given, only nonexistent pathnames that have write permission are shown.\n");
			return 0;
		}
	}
	while (memset(buffer, 0, sizeof(buffer)), fscanf(stdin, "%16380s", buffer) == 1) {
		static int skip_next = 0;
		if (detect_all == 0) {
			unsigned int perm;
			/*
			 * Strictly speaking, I have to ignore the following 2 cases.
			 *
			 *   Domain definition:  <domain> /sbin/init /some_dir/some_program
			 *   Signal definition:  allow_signal 2 <domain> /sbin/init /some_dir/some_program
			 *
			 * But if /some_dir/some_program doesn't exist,
			 * the domain defined will never be used or the signal ACLs given will never be used.
			 * So I will display such pathnames anyway, even though the user wants to display
			 * only nonexistent pathnames that have write permission.
			 */
			if (sscanf(buffer, "%u", &perm) == 1 && (perm & 2) == 0) {
				skip_next = 1;
				continue;
			}
		}
		if (skip_next) {
			skip_next = 0;
			continue;
		}
		if (buffer[0] != '/') continue;
		{
			static char buffer2[sizeof(buffer)];
			struct stat buf;
			if (!decode(buffer, buffer2)) continue;
			if (lstat(buffer2, &buf) == 0) continue;
		}
		for (i = 0; i < pattern_list_count; i++) {
			if (strcmp(pattern_list[i], buffer) == 0) break;
		}
		if (i < pattern_list_count) continue;
		if ((pattern_list = (const char **) realloc(pattern_list, sizeof(const char *) * (pattern_list_count + 1))) == NULL ||
			(pattern_list[pattern_list_count++] = strdup(buffer)) == NULL) {
			fprintf(stderr, "Out of memory.\n");
			exit(1);
		}
	}
	for (i = 0; i < pattern_list_count; i++) {
		for (j = i + 1; j < pattern_list_count; j++) {
			if (strcmp(pattern_list[i], pattern_list[j]) > 0) {
				const char *tmp = pattern_list[i]; pattern_list[i] = pattern_list[j]; pattern_list[j] = tmp;
			}
		}
	}
	for (i = 0; i < pattern_list_count; i++) printf("%s\n", pattern_list[i]);
	return 0;
}
